/**
 * Copyright (c) 2000-present Liferay, Inc. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version.
 *
 * This library is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
 * details.
 */

package com.liferay.portal.search.solr7.internal.search.engine.adapter.search;

import com.liferay.portal.kernel.search.Field;
import com.liferay.portal.kernel.search.GroupBy;
import com.liferay.portal.kernel.search.Stats;
import com.liferay.portal.kernel.search.highlight.HighlightUtil;
import com.liferay.portal.kernel.util.ArrayUtil;
import com.liferay.portal.kernel.util.ListUtil;
import com.liferay.portal.kernel.util.MapUtil;
import com.liferay.portal.kernel.util.SetUtil;
import com.liferay.portal.search.engine.adapter.search.SearchSearchRequest;
import com.liferay.portal.search.groupby.GroupByRequest;
import com.liferay.portal.search.legacy.groupby.GroupByRequestFactory;
import com.liferay.portal.search.legacy.stats.StatsRequestBuilderFactory;
import com.liferay.portal.search.solr7.internal.groupby.GroupByTranslator;
import com.liferay.portal.search.solr7.internal.stats.StatsTranslator;
import com.liferay.portal.search.sort.Sort;
import com.liferay.portal.search.sort.SortFieldTranslator;
import com.liferay.portal.search.stats.StatsRequestBuilder;

import java.util.HashSet;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Objects;
import java.util.Set;
import java.util.stream.Stream;

import org.apache.solr.client.solrj.SolrQuery;

import org.osgi.service.component.annotations.Component;
import org.osgi.service.component.annotations.Reference;

/**
 * @author Bryan Engler
 */
@Component(immediate = true, service = SearchSolrQueryAssembler.class)
public class SearchSolrQueryAssemblerImpl implements SearchSolrQueryAssembler {

	@Override
	public void assemble(
		SolrQuery solrQuery, SearchSearchRequest searchSearchRequest) {

		_baseSolrQueryAssembler.assemble(solrQuery, searchSearchRequest);

		setGroupBy(solrQuery, searchSearchRequest);
		setGroupByRequests(solrQuery, searchSearchRequest);
		setHighlights(solrQuery, searchSearchRequest);
		setIncludeScore(solrQuery, searchSearchRequest);
		setPagination(solrQuery, searchSearchRequest);
		setSelectedFields(solrQuery, searchSearchRequest);
		setSorts(solrQuery, searchSearchRequest);
		setStats(solrQuery, searchSearchRequest);
	}

	protected void addHighlightedField(
		SolrQuery solrQuery, Locale locale, String fieldName) {

		solrQuery.addHighlightField(fieldName);

		String localizedFieldName = Field.getLocalizedName(locale, fieldName);

		solrQuery.addHighlightField(localizedFieldName);
	}

	protected String getSortFieldName(
		com.liferay.portal.kernel.search.Sort sort, String scoreFieldName) {

		String sortFieldName = sort.getFieldName();

		if (Objects.equals(sortFieldName, Field.PRIORITY)) {
			return sortFieldName;
		}

		return Field.getSortFieldName(sort, scoreFieldName);
	}

	@Reference(unbind = "-")
	protected void setBaseSolrQueryAssembler(
		BaseSolrQueryAssembler baseSolrQueryAssembler) {

		_baseSolrQueryAssembler = baseSolrQueryAssembler;
	}

	protected void setGroupBy(
		SolrQuery solrQuery, SearchSearchRequest searchSearchRequest) {

		GroupBy groupBy = searchSearchRequest.getGroupBy();

		if (groupBy != null) {
			_groupByTranslator.translate(
				solrQuery, _groupByRequestFactory.getGroupByRequest(groupBy),
				searchSearchRequest.getLocale(),
				searchSearchRequest.getHighlightFieldNames(),
				searchSearchRequest.isHighlightEnabled(),
				searchSearchRequest.isHighlightRequireFieldMatch(),
				searchSearchRequest.getHighlightFragmentSize(),
				searchSearchRequest.getHighlightSnippetSize());
		}
	}

	@Reference(unbind = "-")
	protected void setGroupByRequestFactory(
		GroupByRequestFactory groupByRequestFactory) {

		_groupByRequestFactory = groupByRequestFactory;
	}

	protected void setGroupByRequests(
		SolrQuery solrQuery, SearchSearchRequest searchSearchRequest) {

		List<GroupByRequest> groupByRequests =
			searchSearchRequest.getGroupByRequests();

		if (ListUtil.isNotEmpty(groupByRequests)) {
			_groupByTranslator.translate(
				solrQuery, groupByRequests.get(0),
				searchSearchRequest.getLocale(),
				searchSearchRequest.getHighlightFieldNames(),
				searchSearchRequest.isHighlightEnabled(),
				searchSearchRequest.isHighlightRequireFieldMatch(),
				searchSearchRequest.getHighlightFragmentSize(),
				searchSearchRequest.getHighlightSnippetSize());
		}
	}

	@Reference(unbind = "-")
	protected void setGroupByTranslator(GroupByTranslator groupByTranslator) {
		_groupByTranslator = groupByTranslator;
	}

	protected void setHighlights(
		SolrQuery solrQuery, SearchSearchRequest searchSearchRequest) {

		if (!searchSearchRequest.isHighlightEnabled()) {
			return;
		}

		solrQuery.setHighlight(true);
		solrQuery.setHighlightFragsize(
			searchSearchRequest.getHighlightFragmentSize());
		solrQuery.setHighlightSimplePost(HighlightUtil.HIGHLIGHT_TAG_CLOSE);
		solrQuery.setHighlightSimplePre(HighlightUtil.HIGHLIGHT_TAG_OPEN);
		solrQuery.setHighlightSnippets(
			searchSearchRequest.getHighlightSnippetSize());

		for (String highlightFieldName :
				searchSearchRequest.getHighlightFieldNames()) {

			addHighlightedField(
				solrQuery, searchSearchRequest.getLocale(), highlightFieldName);
		}

		if (!searchSearchRequest.isLuceneSyntax()) {
			solrQuery.setHighlightRequireFieldMatch(
				searchSearchRequest.isHighlightRequireFieldMatch());
		}
	}

	protected void setIncludeScore(
		SolrQuery solrQuery, SearchSearchRequest searchSearchRequest) {

		solrQuery.setIncludeScore(searchSearchRequest.isScoreEnabled());
	}

	protected void setPagination(
		SolrQuery solrQuery, SearchSearchRequest searchSearchRequest) {

		GroupBy groupBy = searchSearchRequest.getGroupBy();

		if ((groupBy != null) ||
			ListUtil.isNotEmpty(searchSearchRequest.getGroupByRequests())) {

			return;
		}

		solrQuery.setRows(searchSearchRequest.getSize());
		solrQuery.setStart(searchSearchRequest.getStart());
	}

	protected void setSelectedFields(
		SolrQuery solrQuery, SearchSearchRequest searchSearchRequest) {

		if (searchSearchRequest.isAllFieldsSelected()) {
			return;
		}

		Set<String> selectedFieldNames = SetUtil.fromArray(
			searchSearchRequest.getSelectedFieldNames());

		if (!selectedFieldNames.contains(Field.UID)) {
			selectedFieldNames.add(Field.UID);
		}

		solrQuery.setFields(selectedFieldNames.toArray(new String[0]));
	}

	@Reference(unbind = "-")
	protected void setSortFieldTranslator(
		SortFieldTranslator<SolrQuery.SortClause> sortFieldTranslator) {

		_sortFieldTranslator = sortFieldTranslator;
	}

	protected void setSorts(
		SolrQuery solrQuery, SearchSearchRequest searchSearchRequest) {

		List<Sort> sorts = searchSearchRequest.getSorts();

		Stream<Sort> stream = sorts.stream();

		stream.map(
			_sortFieldTranslator::translate
		).forEach(
			solrQuery::addSort
		);

		com.liferay.portal.kernel.search.Sort[] sorts71 =
			searchSearchRequest.getSorts71();

		if (ArrayUtil.isEmpty(sorts71)) {
			return;
		}

		Set<String> sortFieldNames = new HashSet<>(sorts71.length);

		for (com.liferay.portal.kernel.search.Sort sort : sorts71) {
			if (sort == null) {
				continue;
			}

			String sortFieldName = getSortFieldName(sort, "score");

			if (sortFieldNames.contains(sortFieldName)) {
				continue;
			}

			sortFieldNames.add(sortFieldName);

			SolrQuery.ORDER order = SolrQuery.ORDER.asc;

			if (sort.isReverse() || sortFieldName.equals("score")) {
				order = SolrQuery.ORDER.desc;
			}

			solrQuery.addSort(new SolrQuery.SortClause(sortFieldName, order));
		}
	}

	protected void setStats(
		SolrQuery solrQuery, SearchSearchRequest searchSearchRequest) {

		Map<String, Stats> statsMap = searchSearchRequest.getStats();

		if (!MapUtil.isEmpty(statsMap)) {
			for (Stats stats : statsMap.values()) {
				StatsRequestBuilder statsRequestBuilder =
					_statsRequestBuilderFactory.getStatsRequestBuilder(stats);

				_statsTranslator.populateRequest(
					solrQuery, statsRequestBuilder.build());
			}
		}
	}

	@Reference(unbind = "-")
	protected void setStatsRequestBuilderFactory(
		StatsRequestBuilderFactory statsRequestBuilderFactory) {

		_statsRequestBuilderFactory = statsRequestBuilderFactory;
	}

	@Reference(unbind = "-")
	protected void setStatsTranslator(StatsTranslator statsTranslator) {
		_statsTranslator = statsTranslator;
	}

	private BaseSolrQueryAssembler _baseSolrQueryAssembler;
	private GroupByRequestFactory _groupByRequestFactory;
	private GroupByTranslator _groupByTranslator;
	private SortFieldTranslator<SolrQuery.SortClause> _sortFieldTranslator;
	private StatsRequestBuilderFactory _statsRequestBuilderFactory;
	private StatsTranslator _statsTranslator;

}